type CaptureSet = set[str] | dict[str, str]
"""A set of captured or cloned variables"""


def normalize_capture_set(capture: CaptureSet | None) -> dict[str, str]:
    """Makes sure the capture set is a dict"""

    match capture:
        case None:
            capture = {}
        case set():
            capture = {n: n for n in capture}
        case dict():
            pass
        case _:
            raise TypeError("Invalid capture set")
    return capture



import inspect
from collections.abc import Callable
from enum import Enum, Flag, auto
from functools import wraps
from types import TracebackType
from typing import Any, Self, Literal

from pypactum._evaluation_semantic import EvaluationSemantic
from pypactum._assertion_kind import AssertionKind
from pypactum._utils._assert_contract import assert_contract
from pypactum._utils._effective_semantic import effective_semantic
from pypactum._utils._map_function_arguments import map_function_arguments
from pypactum._utils._parent_frame import get_parent_frame
from pypactum._utils._resolve_bindings import (
    resolve_bindings,
    collect_available_variables,
)
from pypactum._predicate import Predicate
from pypactum._capture_set import CaptureSet, normalize_capture_set
from pypactum._contract_assertion_label import ContractAssertionLabel


class PostconditionScope(Flag):
    """Holds information on when a postcondition should be evaluated."""

    RegularReturn = auto()
    ExceptionalReturn = auto()
    Always = RegularReturn | ExceptionalReturn


class post:
    """Postcondition assertion factory taking a predicate to evaluate after function evaluation"""

    def __init__(
        self,
        predicate: Predicate,
        /,
        *,
        capture_before: CaptureSet | None = None,
        capture_after: CaptureSet | None = None,
        clone_before: CaptureSet | None = None,
        clone_after: CaptureSet | None = None,
        labels: list[ContractAssertionLabel] | None = None,
        scope: PostconditionScope = PostconditionScope.RegularReturn,
    ):
        """Initializes the precondition assertion factory

        Keyword arguments:
            predicate: A callable evaluating the predicate to check after function evaluation
            capture_before: A set of names to capture before function evaluation. Variables by this name can be predicate parameters
            capture_after: A set of names to capture after function evaluation. Variables by this name can be predicate parameters
            clone_before: A set of names to clone before function evaluation. Variables by this name can be predicate parameters
            clone_after: A set of names to clone after function evaluation. Variables by this name can be predicate parameters
            labels: A list of labels that determine this assertion's evaluation semantic
            scope: Determines whether the postcondition applies if the function returns regularly, with an exception, or both
        """

        capture_before = normalize_capture_set(capture_before)
        capture_after = normalize_capture_set(capture_after)
        clone_before = normalize_capture_set(clone_before)
        clone_after = normalize_capture_set(clone_after)

        if labels is None:
            labels = []

        self.__predicate = predicate
        self.__capture_before = capture_before
        self.__capture_after = capture_after
        self.__clone_before = clone_before
        self.__clone_after = clone_after
        self.__scope = scope
        self.__parent_frame = get_parent_frame(inspect.currentframe())
        self.__semantic = effective_semantic(
            self.__parent_frame, AssertionKind.post, labels
        )

    def __find_result_param(self) -> str | None:
        """Given the predicate parameters `pred_params`, finds the one not in the set of bound names `bindings`.

        Returns `None` if there is no result parameter.

        Raises `TypeError` if there is more than one potential result parameter.
        """

        bindings = (
            set(self.__capture_before.keys())
            | set(self.__clone_before.keys())
            | set(self.__capture_after.keys())
            | set(self.__clone_after.keys())
        )
        params = inspect.signature(self.__predicate).parameters
        candidates = {n for n in params.keys() if n not in bindings}
        match len(candidates):
            case 0:
                return None
            case 1:
                return candidates.pop()
            case _:
                raise TypeError(
                    f"Unable to determine predicate result parameter. Candidates: {','.join(candidates)}"
                )

    def __call__[R](
        self,
        func: Callable[..., R],
        /,
        *,
        _implicit_return_capture: bool = True,
        _implicit_arg_capture: bool = False,
    ) -> Callable[..., R]:
        """Wraps the given callable in another callable that checks postconditions after executing the original callable

        Keyword arguments:
            func: Callable to wrap. Typically, a function.

        Returns:
            - `func` if the effective contract evaluation semantic is `ignore`
            - a checked wrapper compatible with `func` otherwise
        """

        if self.__semantic == EvaluationSemantic.ignore:
            return func

        @wraps(func)
        def checked_func(*args: Any, **kwargs: Any) -> R:
            sig = inspect.signature(func)
            nkwargs = map_function_arguments(sig, args, kwargs)

            # resolve "before"-type bindings
            available_variables = collect_available_variables(
                self.__parent_frame,
                nkwargs,
            )
            capture = self.__capture_before
            if _implicit_arg_capture:
                capture = {n: n for n in sig.parameters.keys()} | capture
            resolved_kwargs = resolve_bindings(
                available_variables=available_variables,
                capture=capture,
                clone=self.__clone_before,
            )

            # evaluate decorated function
            exception_raised = None
            try:
                result = func(*args, **kwargs)
            except Exception as exc:
                if PostconditionScope.ExceptionalReturn not in self.__scope:
                    raise
                exception_raised = exc
            else:
                if PostconditionScope.RegularReturn not in self.__scope:
                    return result

            # resolve "after"-type bindings
            available_variables = collect_available_variables(
                self.__parent_frame,
                nkwargs,
            )
            # Implicitly capture result argument
            capture = self.__capture_after
            if _implicit_return_capture:
                result_name = self.__find_result_param()
                result_value = result if exception_raised is None else exception_raised
                if result_name is not None:
                    capture = {result_name: result_name} | capture
                    available_variables.insert(0, {result_name: result_value})
            resolved_kwargs |= resolve_bindings(
                available_variables=available_variables,
                capture=capture,
                clone=self.__clone_after,
            )

            # assert postcondition
            assert_contract(
                semantic=self.__semantic,
                kind=AssertionKind.post,
                calling_frame=self.__parent_frame,
                predicate=self.__predicate,
                predicate_kwargs=resolved_kwargs,
            )

            # If an exception was raised, re-raise it; otherwise, return the regular return value
            if exception_raised is not None:
                raise exception_raised
            else:
                return result

        return checked_func

    def __enter__(self) -> Self:
        """Captures before-type bindings when the scope is entered"""

        # resolve "before"-type bindings
        available_variables = collect_available_variables(self.__parent_frame, {})
        self.__resolved_kwargs = resolve_bindings(
            available_variables=available_variables,
            capture=self.__capture_before,
            clone=self.__clone_before,
        )
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        """Captures after-type bindings, then checks the postcondition"""

        exceptional_exit = exc_val is not None
        if exceptional_exit:
            if PostconditionScope.ExceptionalReturn not in self.__scope:
                return False
        else:
            if PostconditionScope.RegularReturn not in self.__scope:
                return False

        # resolve "after"-type bindings
        available_variables = collect_available_variables(self.__parent_frame, {})
        self.__resolved_kwargs |= resolve_bindings(
            available_variables=available_variables,
            capture=self.__capture_after,
            clone=self.__clone_after,
        )

        # assert postcondition
        assert_contract(
            semantic=self.__semantic,
            kind=AssertionKind.post,
            calling_frame=self.__parent_frame,
            predicate=self.__predicate,
            predicate_kwargs=self.__resolved_kwargs,
        )
        return False



import pytest

from pypactum import (
    ContractViolationException,
    pre,
    post,
    labels,
    global_contract_assertion_label,
    PostconditionScope,
)

THE_ANSWER = [42]


def __raise(exception):
    raise exception


def test_post_predicate_false():

    @pre(lambda: True)
    @post(lambda: False)
    def test():
        pass

    with pytest.raises(ContractViolationException):
        test()


def test_post_predicate_true():

    @post(lambda: True)
    def test():
        return 42

    assert test() == 42


def test_post_predicate_with_result_argument():

    @post(lambda result: result <= 50)
    def test(x, y):
        return x + y

    assert test(1, 11) == 12
    assert test(20, 30) == 50

    with pytest.raises(ContractViolationException):
        test(50, 49)


def test_post_predicate_with_capture():

    @post(lambda result, x, y: result == x + y, capture_before={"x", "y"})
    def test(x, y):
        return abs(x + y)

    assert test(1, 11) == 12
    assert test(20, 30) == 50

    @post(lambda x, y, result: result == x + y, capture_before={"x", "y"})
    def test(x, y):
        return abs(x + y)

    assert test(1, 11) == 12
    assert test(20, 30) == 50

    with pytest.raises(ContractViolationException):
        test(-2, 1)

    @post(lambda result, z, y: True, capture_before={"z", "y"})
    def test(x, y):
        return abs(x + y)

    with pytest.raises(TypeError):
        test(1, 2)

    @post(lambda x: not x.append(0), capture_before={"x"})
    def test(x):
        return len(x) == 0

    assert test([])

    @post(
        lambda args, result, kwargs: result == len(args) + len(kwargs),
        capture_before={"args", "kwargs"},
    )
    def test(*args, **kwargs):
        return len(args) + len(kwargs)

    assert test(1, 2, 3) == 3


def test_post_exception():

    @post(lambda: __raise(ValueError("Ahhhh")))
    def test():
        pass

    with pytest.raises(ValueError):
        test()


def test_post_scope_exceptional_return():

    @post(
        lambda exc: type(exc) == ValueError, scope=PostconditionScope.ExceptionalReturn
    )
    def test(exc):
        raise exc

    with pytest.raises(ValueError):
        test(ValueError())

    with pytest.raises(ContractViolationException):
        test(TypeError())


def test_post_scope_always():

    @post(lambda r: type(r) in [ValueError, int], scope=PostconditionScope.Always)
    def test(val):
        if isinstance(val, Exception):
            raise val
        return val

    test(42)

    with pytest.raises(ValueError):
        test(ValueError())

    with pytest.raises(ContractViolationException):
        test(TypeError())

    with pytest.raises(ContractViolationException):
        test("something")


def test_post_predicate_invalid():

    @post(lambda x, y: True)
    def test():
        pass

    with pytest.raises(TypeError):
        test()


def test_post_capture_before_local():

    @post(lambda x: x == 0, capture_before={"x"})
    def test():
        pass

    x = 0

    test()


def test_post_clone_before_local():

    @post(lambda x: x.pop() == 0, clone_before={"x"})
    def test():
        pass

    x = [0]

    test()

    assert x == [0]


def test_post_capture_before_global():

    @post(lambda THE_ANSWER: THE_ANSWER[0] == 42, capture_before={"THE_ANSWER"})
    def test():
        pass

    test()


def test_post_clone_before_global():

    @post(lambda THE_ANSWER: THE_ANSWER.pop() == 42, clone_before={"THE_ANSWER"})
    def test():
        pass

    test()

    assert THE_ANSWER == [42]


def test_post_capture_after_local():

    @post(lambda x: x == 0, capture_after={"x"})
    def test():
        nonlocal x
        x = 0

    x = 42

    test()

    assert x == 0


def test_post_clone_after_local():

    @post(lambda x: x.pop() == 1, clone_after={"x"})
    def test():
        nonlocal x
        x[0] = 1

    x = [42]

    test()

    assert x == [1]


def test_post_capture_after_global():

    @post(lambda THE_ANSWER: THE_ANSWER[0] == 42, capture_after={"THE_ANSWER"})
    def test():
        pass

    test()


def test_post_clone_after_global():

    @post(lambda THE_ANSWER: THE_ANSWER.pop() == 42, clone_after={"THE_ANSWER"})
    def test():
        pass

    test()

    assert THE_ANSWER == [42]


def test_post_capture_rename():

    @post(lambda y: y == 0, capture_before={"y": "x"})
    def test():
        pass

    x = 0

    test()


def test_post_capture_before_and_after():

    @post(lambda x, y: len(x) + 1 == len(y), clone_before={"x"}, clone_after={"y": "x"})
    def test(x):
        x.append(42)

    test([42])


def test_post_capture_wrong_type():

    with pytest.raises(TypeError):

        @post(lambda x: x == 0, capture_before=["x"])
        def test():
            pass


def test_post_as_context_manager():
    x = [42]
    with (
        post(lambda x: x.pop() == 42, clone_before={"x"}),
        post(lambda x: x.pop() == 3, clone_after={"x"}),
    ):
        x[0] = 3
    assert x == [3]


def test_post_as_context_manager_scope_regular_exit():
    x = [42]

    with pytest.raises(ValueError):
        with post(lambda x: x[0] == 0, capture_after={"x"}):
            raise ValueError()


def test_post_as_context_manager_scope_exceptional_exit():
    x = [42]

    with post(
        lambda x: x[0] == 0,
        capture_after={"x"},
        scope=PostconditionScope.ExceptionalReturn,
    ):
        pass


def test_post_with_label():

    @post(lambda: False, labels=[labels.ignore])
    def test():
        pass

    test()

    @post(lambda: False, labels=[labels.ignore_postconditions])
    def test():
        pass

    test()

    with labels.enable_expensive(False):

        @post(lambda: False, labels=[labels.expensive])
        def test():
            pass

        test()

    with pytest.raises(ContractViolationException), labels.enable_expensive(True):

        @post(lambda: False, labels=[labels.expensive])
        def test():
            pass

        test()

    with global_contract_assertion_label(labels.filter_by_module(r"foo")):

        @post(lambda: False)
        def test():
            pass

        test()

    with global_contract_assertion_label(labels.filter_by_module(r"post")):

        with pytest.raises(ContractViolationException):

            @post(lambda: False)
            def test():
                pass

            test()



import logging

import pytest

from pypactum import (
    ContractViolationException,
    contract_violation_handler,
    handlers,
    pre,
)


def test_raising_handler():
    with (
        pytest.raises(ContractViolationException),
        contract_violation_handler(handlers.raise_on_contract_violation),
    ):

        @pre(lambda: False)
        def test():
            pass

        test()


def test_logging_handler_stderr(capsys):
    with contract_violation_handler(handlers.log_on_contract_violation()):

        @pre(lambda: False)
        def test():
            pass

        test()

        out, err = capsys.readouterr()
        assert "Precondition violation" in err


def test_logging_handler_logging(caplog):
    with contract_violation_handler(handlers.log_on_contract_violation(logging.root)):

        @pre(lambda: False)
        def test():
            pass

        test()

        assert any("Precondition violation" in l.message for l in caplog.records)



import sys
import logging
from collections.abc import Callable
from io import TextIOWrapper
from typing import TextIO

from pypactum._contract_violation import ContractViolation
from pypactum._contract_violation_exception import ContractViolationException

type ContractViolationHandler = Callable[[ContractViolation], None]


def raise_on_contract_violation(violation: ContractViolation) -> None:
    """A contract violation handler that raises a ContractViolationException"""
    raise ContractViolationException(violation)


def log_on_contract_violation(
    target: logging.Logger | TextIO | TextIOWrapper | None = None,
) -> ContractViolationHandler:
    """A factory for contract violation handlers that log the violation

    Keyword arguments:
        target: Either a TextIO sink like stderr (the default), or a logger object
    """
    if target is None:
        target = sys.stderr

    match target:
        case TextIO() | TextIOWrapper() as file:

            def handler(violation: ContractViolation) -> None:
                print(str(violation), file=file)

            return handler

        case logging.Logger() as log:

            def handler(violation: ContractViolation) -> None:
                log.error(str(violation))

            return handler

        case _:
            raise TypeError("Invalid logging target")



from warnings import deprecated

import pytest

from pypactum import (
    ContractViolationException,
    pre,
    post,
    labels,
    global_contract_assertion_label,
)

THE_ANSWER = [42]


def __raise(exception):
    raise exception


def test_pre_predicate_false():

    @pre(lambda: False)
    def test():
        pass

    with pytest.raises(ContractViolationException):
        test()


def test_pre_predicate_true():

    @pre(lambda: True)
    def test():
        return 42

    assert test() == 42


def test_pre_predicate_with_named_arguments():

    @pre(lambda x: x > 0)
    @pre(lambda y: y > 10)
    @pre(lambda y, x: x + y < 100)
    def test(x, y):
        return x + y

    assert test(1, 11) == 12
    assert test(50, 49) == 99

    with pytest.raises(ContractViolationException):
        test(0, 20)

    with pytest.raises(ContractViolationException):
        test(20, 0)

    with pytest.raises(ContractViolationException):
        test(200, 200)


def test_pre_predicate_with_args():

    @pre(lambda args: len(args) > 0)
    def test(*args):
        return args[0]

    assert test(42) == 42
    assert test(42, 19) == 42

    with pytest.raises(ContractViolationException):
        test()

    @pre(lambda args: len(args) > 0)
    def test(*args):
        return args[0]

    assert test(42) == 42
    assert test(42, 19) == 42


def test_pre_predicate_with_kwargs():

    @pre(lambda kwargs: "y" in kwargs)
    def test(x, **kwargs):
        return x

    assert test(x=42, y=1) == 42
    assert test(x=42, y="foobar", z=3) == 42

    with pytest.raises(ContractViolationException):
        test(x=42)

    with pytest.raises(TypeError):
        test(y=9)


def test_pre_predicate_with_mixed_arguments():

    @pre(lambda x: x > 0)
    @pre(lambda y: y > 10)
    @pre(lambda y, x: x + y < 100)
    @pre(lambda args, x, kwargs: True)
    @pre(lambda args: len(args) > 0)
    @pre(lambda kwargs: len(kwargs) >= 1)
    def test(x, /, y, *args, z=0, **kwargs):
        return x + y + z

    assert test(1, 11, 42, 43, foo=9) == 12
    assert test(50, 49, 42, 43, foo=9) == 99

    with pytest.raises(ContractViolationException):
        test(0, 20, 42, 43, foo=9)

    with pytest.raises(ContractViolationException):
        test(20, 0, 42, 43, foo=9)

    with pytest.raises(ContractViolationException):
        test(200, 200, 42, 43, foo=9)

    with pytest.raises(ContractViolationException):
        test(50, 49, 1)

    with pytest.raises(ContractViolationException):
        test(50, 49, foo=1)


def test_pre_exception():

    @pre(lambda: __raise(ValueError("Ahhhh")))
    def test():
        pass

    with pytest.raises(ValueError):
        test()


def test_pre_predicate_invalid():

    @pre(lambda x: x == 0)
    def test():
        pass

    with pytest.raises(TypeError):
        test()


def test_pre_explicit_capture_has_precedence():

    @pre(lambda x: x == 0, capture={"x": "y"})
    def test(x, y):
        pass

    test(1, 0)

    with pytest.raises(ContractViolationException):
        test(0, 1)


def test_pre_capture_local():

    @pre(lambda x: x == 0, capture={"x"})
    def test():
        pass

    x = 0

    test()


def test_pre_clone_local():

    @pre(lambda x: x.pop() == 0, clone={"x"})
    def test():
        pass

    x = [0]

    test()

    assert x == [0]


def test_pre_capture_global():

    @pre(lambda THE_ANSWER: THE_ANSWER[0] == 42, capture={"THE_ANSWER"})
    def test():
        pass

    test()


def test_pre_clone_global():

    @pre(lambda THE_ANSWER: THE_ANSWER.pop() == 42, clone={"THE_ANSWER"})
    def test():
        pass

    test()

    assert THE_ANSWER == [42]


def test_pre_capture_rename():

    @pre(lambda y: y == 0, capture={"y": "x"})
    def test():
        pass

    x = 0

    test()


def test_pre_post_capture_with_other_decorator():

    x = ["foo"]

    @deprecated("foobar")
    @pre(lambda x: len(x) == 1, capture={"x"})
    @post(lambda result, x, y: len(result) == len(x) + 1, capture_before={"x", "y"})
    def test(y: int):
        return x + [y]

    test(42)


def test_pre_capture_wrong_type():

    with pytest.raises(TypeError):

        @pre(lambda x: x == 0, capture=["x"])
        def test():
            pass


def test_pre_as_context_manager():
    x = [42]
    with pre(lambda x: x.pop() == 42, clone={"x"}):
        pass
    assert x == [42]


def test_pre_post_as_context_manager():
    x = [42]
    with (
        pre(lambda x: x.pop() == 42, clone={"x"}),
        post(lambda x: len(x) == 0, capture_after={"x"}),
    ):
        x.pop()
    assert x == []


def test_pre_with_label():

    @pre(lambda: False, labels=[labels.ignore])
    def test():
        pass

    test()

    with pytest.raises(ContractViolationException):

        @pre(lambda: False, labels=[labels.ignore_postconditions])
        def test():
            pass

        test()

    with labels.enable_expensive(False):

        @pre(lambda: False, labels=[labels.expensive])
        def test():
            pass

        test()

    with pytest.raises(ContractViolationException), labels.enable_expensive(True):

        @pre(lambda: False, labels=[labels.expensive])
        def test():
            pass

        test()

    with global_contract_assertion_label(labels.filter_by_module(r"foo")):

        @pre(lambda: False)
        def test():
            pass

        test()

    with global_contract_assertion_label(labels.filter_by_module(r"pre")):

        with pytest.raises(ContractViolationException):

            @pre(lambda: False)
            def test():
                pass

            test()



import inspect
from typing import Any


def map_function_arguments(
    signature: inspect.Signature,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> dict[str, Any]:
    """Maps actual function arguments from `args` and `kwargs` to their declared names given `signature`."""

    bound_args = signature.bind(*args, **kwargs)
    bound_args.apply_defaults()
    return dict(bound_args.arguments)



from dataclasses import dataclass

import pytest

from pypactum import (
    ContractViolationException,
    invariant,
)


def __raise(exception):
    raise exception


def test_invariant_predicate_false():

    @invariant(lambda: False)
    class Test:
        pass

    with pytest.raises(ContractViolationException):
        Test()


def test_invariant_predicate_false_after_init():

    @invariant(lambda self: self.x < 0)
    class Test:
        def __init__(self):
            self.x = 42

    with pytest.raises(ContractViolationException):
        Test()


def test_invariant_predicate_false_before_or_after_method_call():

    @invariant(lambda self: self.x > 0)
    class Test:
        def __init__(self):
            self.x = 42

        def foo(self):
            pass

        def bar(self):
            self.x = -1

    t = Test()
    t.foo()

    t.x = -1  # Break the invariant
    with pytest.raises(ContractViolationException):
        t.foo()

    t.x = 42  # Fix the invariant
    with pytest.raises(ContractViolationException):
        t.bar()  # This breaks the invariant


def test_invariant_broken_after_exception():
    @invariant(lambda self: self.x > 0)
    class Test:
        def __init__(self):
            self.x = 42

        def foo(self, y):
            self.x -= 100
            self.x += 1000 / y  # Raises exception if y==0

    t = Test()
    t.foo(10)

    with pytest.raises(ContractViolationException):
        t.foo(0)


def test_invariant_on_dataclass():
    @invariant(lambda self: self.foo > 0)
    @dataclass
    class Foo:
        foo: int

    f = Foo(42)

    with pytest.raises(ContractViolationException):
        f = Foo(-1)


def test_two_invariants():
    @invariant(lambda self: self.x > 0)
    @invariant(lambda self: self.y < 0)
    class Test:
        def __init__(self, x, y):
            self.x = x
            self.y = y

    t = Test(1, -1)

    with pytest.raises(ContractViolationException):
        Test(0, -1)

    with pytest.raises(ContractViolationException):
        Test(1, 1)


def test_invariant_as_context_manager():

    x = 42
    with invariant(lambda x: x == 42, capture={"x"}):
        x *= 2
        x -= 42



from pypactum._contract_violation import ContractViolation


class ContractViolationException(Exception):
    """Exception to raise in case of contract violation"""

    def __init__(self, violation: ContractViolation):
        self.violation = violation



from contextlib import ContextDecorator
from types import TracebackType
from typing import Self, Literal

from pypactum._contract_violation import ContractViolation
from pypactum._evaluation_semantic import EvaluationSemantic
from pypactum._contract_assertion_label import (
    ContractAssertionLabel,
    ContractAssertionInfo,
)
from pypactum.handlers import raise_on_contract_violation, ContractViolationHandler

__contract_violation_handler: ContractViolationHandler = raise_on_contract_violation
__global_evaluation_semantic: EvaluationSemantic = EvaluationSemantic.check
__global_contract_assertion_label: ContractAssertionLabel = lambda sem, info: sem


def invoke_contract_violation_handler(violation: ContractViolation) -> None:
    """Invokes the current contract violation handler"""
    __contract_violation_handler(violation)


def set_contract_violation_handler(handler: ContractViolationHandler) -> None:
    """Replaces the contract violation handler"""
    global __contract_violation_handler
    __contract_violation_handler = handler


def get_contract_violation_handler() -> ContractViolationHandler:
    """Retrieves the current contract violation handler"""
    return __contract_violation_handler


def set_contract_evaluation_semantic(semantic: EvaluationSemantic) -> None:
    """Changes the current contract evaluation semantic"""
    global __global_evaluation_semantic
    __global_evaluation_semantic = semantic


def get_contract_evaluation_semantic(
    info: ContractAssertionInfo | None = None,
) -> EvaluationSemantic:
    """Retrieves the currently configured contract evaluation semantic"""

    if info is None:
        return __global_evaluation_semantic
    return __global_contract_assertion_label(__global_evaluation_semantic, info)


def set_global_contract_assertion_label(label: ContractAssertionLabel) -> None:
    """Changes the current global label, influencing the evaluation semantics"""
    global __global_contract_assertion_label
    __global_contract_assertion_label = label


def get_global_contract_assertion_label() -> ContractAssertionLabel:
    """Retrieves the current global label"""
    return __global_contract_assertion_label


class contract_violation_handler(ContextDecorator):
    """Context manager to temporarily change the contract violation handler. Usable as decorator as well."""

    def __init__(self, handler: ContractViolationHandler):
        self.handler = handler

    def __enter__(self) -> Self:
        self.old_handler = get_contract_violation_handler()
        set_contract_violation_handler(self.handler)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        set_contract_violation_handler(self.old_handler)
        return False


class contract_evaluation_semantic(ContextDecorator):
    """Context manager to temporarily change the evaluation semantic. Usable as a decorator as well."""

    def __init__(
        self,
        semantic: EvaluationSemantic,
    ):
        self.semantic = semantic

    def __enter__(self) -> Self:
        self.old_semantic = get_contract_evaluation_semantic()
        set_contract_evaluation_semantic(self.semantic)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        set_contract_evaluation_semantic(self.old_semantic)
        return False


class global_contract_assertion_label(ContextDecorator):
    """Context manager to temporarily change the global contract assertion label. Usable as a decorator as well."""

    def __init__(
        self,
        label: ContractAssertionLabel,
    ):
        self.label = label

    def __enter__(self) -> Self:
        self.old_label = get_global_contract_assertion_label()
        set_global_contract_assertion_label(self.label)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        set_global_contract_assertion_label(self.old_label)
        return False



import inspect
from typing import Any

from pypactum._assertion_kind import AssertionKind
from pypactum._evaluation_semantic import EvaluationSemantic


class ContractViolation:
    """Holds information about a contract violation"""

    __kind_strings = [None, "Precondition", "Postcondition", "Assertion"]

    def __init__(
        self,
        comment: str,
        kind: AssertionKind,
        location: inspect.Traceback | None,
        semantic: EvaluationSemantic,
        kwargs: dict[str, Any],
    ):
        self.comment = comment
        self.kind = kind
        self.location = location
        self.semantic = semantic
        self.kwargs = kwargs

    def __str__(self) -> str:
        kind = ContractViolation.__kind_strings[self.kind.value]
        loc = (
            f"{self.location.filename}:{self.location.lineno}"
            if self.location is not None
            else ""
        )
        diagnostic = f"{kind} violation at {loc}"
        if len(self.comment) > 0:
            diagnostic += f": {self.comment}"
        diagnostic += f"\nVariables: {self.kwargs}"
        if (
            self.location is not None
            and self.location.code_context is not None
            and len(self.location.code_context) > 0
        ):
            diagnostic += f"\nContext:\n{'\n'.join(self.location.code_context)}"
        return diagnostic



from types import FrameType


def get_parent_frame(frame: FrameType | None) -> FrameType | None:
    """If `frame` is not None, returns its parent frame. Otherwise, returns None."""

    if frame is not None:
        return frame.f_back
    return frame



import re
from types import TracebackType
from typing import Self, Literal

from pypactum._evaluation_semantic import EvaluationSemantic
from pypactum._assertion_kind import AssertionKind
from pypactum._contract_assertion_label import ContractAssertionInfo


def ignore(
    semantic: EvaluationSemantic,
    info: ContractAssertionInfo,
) -> EvaluationSemantic:
    """A contract assertion label marking everything as ignored"""
    return EvaluationSemantic.ignore


def check(
    semantic: EvaluationSemantic,
    info: ContractAssertionInfo,
) -> EvaluationSemantic:
    """A contract assertion label marking everything as checked"""
    return EvaluationSemantic.check


def ignore_postconditions(
    semantic: EvaluationSemantic,
    info: ContractAssertionInfo,
) -> EvaluationSemantic:
    """A contract assertion label marking all postconditions as ignored"""

    if info.kind == AssertionKind.post:
        return EvaluationSemantic.ignore
    return semantic


__expensive_assertions_enabled = True


def expensive(
    semantic: EvaluationSemantic,
    info: ContractAssertionInfo,
) -> EvaluationSemantic:
    """A contract assertion label marking an assertion as expensive to check"""

    if not __expensive_assertions_enabled:
        return EvaluationSemantic.ignore
    return semantic


def set_expensive_assertions_enabled(enabled: bool) -> None:
    """Allows to enable or disable expensive contract checks"""
    global __expensive_assertions_enabled
    __expensive_assertions_enabled = enabled


def get_expensive_assertions_enabled() -> bool:
    """Retrieves whether expensive contract checks are currently enabled"""
    return __expensive_assertions_enabled


class enable_expensive:
    def __init__(self, enable: bool = True):
        self.enable = enable

    def __enter__(self) -> Self:
        self.__old_enable = get_expensive_assertions_enabled()
        set_expensive_assertions_enabled(self.enable)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        set_expensive_assertions_enabled(self.__old_enable)
        return False


class filter_by_module:
    """A contract assertion label factory that marks assertions as ignored if they don't pass a regex"""

    def __init__(self, regex: re.Pattern[str] | str):
        if isinstance(regex, str):
            self.regex = re.compile(regex)
        else:
            self.regex = regex

    def __call__(
        self,
        semantic: EvaluationSemantic,
        info: ContractAssertionInfo,
    ) -> EvaluationSemantic:
        if self.regex.search(info.module_name) is None:
            return EvaluationSemantic.ignore
        return semantic



from ._assertion_kind import AssertionKind  # noqa: F401
from ._contract_assertion_label import (
    ContractAssertionInfo,  # noqa: F401
    ContractAssertionLabel,  # noqa: F401
)
from ._contract_violation import ContractViolation  # noqa: F401
from ._contract_violation_exception import ContractViolationException  # noqa: F401
from ._contract_violation_handler import (
    get_contract_violation_handler,  # noqa: F401
    set_contract_violation_handler,  # noqa: F401
    contract_violation_handler,  # noqa: F401
    get_contract_evaluation_semantic,  # noqa: F401
    set_contract_evaluation_semantic,  # noqa: F401
    contract_evaluation_semantic,  # noqa: F401
    get_global_contract_assertion_label,  # noqa: F401
    set_global_contract_assertion_label,  # noqa: F401
    global_contract_assertion_label,  # noqa: F401
)
from ._evaluation_semantic import EvaluationSemantic  # noqa: F401
from .handlers import ContractViolationHandler  # noqa: F401
from ._post import post, PostconditionScope  # noqa: F401
from ._pre import pre  # noqa: F401
from ._predicate import Predicate  # noqa: F401
from ._invariant import invariant  # noqa: F401

__all__ = [
    "AssertionKind",
    "ContractAssertionInfo",
    "ContractAssertionLabel",
    "ContractViolation",
    "ContractViolationException",
    "ContractViolationHandler",
    "EvaluationSemantic",
    "PostconditionScope",
    "Predicate",
    "contract_evaluation_semantic",
    "contract_violation_handler",
    "get_contract_evaluation_semantic",
    "get_contract_violation_handler",
    "get_global_contract_assertion_label",
    "global_contract_assertion_label",
    "invariant",
    "post",
    "pre",
    "set_contract_evaluation_semantic",
    "set_contract_violation_handler",
    "set_global_contract_assertion_label",
]



from collections.abc import Callable
from dataclasses import dataclass

from pypactum._evaluation_semantic import EvaluationSemantic
from pypactum._assertion_kind import AssertionKind


@dataclass
class ContractAssertionInfo:
    """Bundles information about a contract assertion relevant to contract assertion labels"""

    kind: AssertionKind
    module_name: str


type ContractAssertionLabel = Callable[
    [EvaluationSemantic, ContractAssertionInfo], EvaluationSemantic
]
"""A contract assertion label is a function that takes the previous evaluation semantic and information about the 
current contract assertion, and returns a new evaluation semantic.
"""



import copy
from types import FrameType
from typing import Any


def __resolve_binding(available_variables: list[dict[str, Any]], name: str) -> Any:
    """Resolves a single binding and returns it. In case of error, raises TypeError."""
    for scope in available_variables:
        if name in scope:
            return scope[name]
    raise TypeError(f'Invalid binding "{name}"')


def resolve_bindings(
    available_variables: list[dict[str, Any]],
    capture: dict[str, str],
    clone: dict[str, str],
) -> dict[str, Any]:
    """Resolves all captures and clones and returns them. In case of error, raises TypeError."""

    referenced = {
        k: __resolve_binding(available_variables, v) for k, v in capture.items()
    }
    cloned = {
        k: copy.deepcopy(__resolve_binding(available_variables, v))
        for k, v in clone.items()
    }
    return referenced | cloned


def collect_available_variables(
    parent_frame: FrameType | None,
    kwargs: dict[str, Any],
) -> list[dict[str, Any]]:
    """Collects a list of all available variables, to be passed to resolve_bindings"""

    candidate_bindings = [kwargs]
    if parent_frame is not None:
        candidate_bindings += [
            parent_frame.f_locals,
            parent_frame.f_globals,
        ]
    return candidate_bindings



import inspect
from types import FrameType
from typing import Any
from pypactum._evaluation_semantic import EvaluationSemantic
from pypactum._assertion_kind import AssertionKind
from pypactum._contract_violation import ContractViolation
from pypactum._contract_violation_handler import invoke_contract_violation_handler
from pypactum._predicate import Predicate


def __handle_contract_violation(
    semantic: EvaluationSemantic,
    kind: AssertionKind,
    location: inspect.Traceback | None,
    kwargs: dict[str, Any],
    comment: str = "",
) -> None:
    """Handles a contract violation by invoking the contract violation handler and/or terminating if required"""

    violation = ContractViolation(
        comment=comment,
        kind=kind,
        location=location,
        semantic=semantic,
        kwargs=kwargs,
    )

    if semantic == EvaluationSemantic.check:
        invoke_contract_violation_handler(violation)


def assert_contract(
    semantic: EvaluationSemantic,
    kind: AssertionKind,
    calling_frame: FrameType | None,
    predicate: Predicate,
    predicate_kwargs: dict[str, Any],
) -> None:
    """Evaluates the given predicate and handles a contract violation if the result was false"""

    if semantic == EvaluationSemantic.ignore:
        return

    kwargs = {
        k: v
        for k, v in predicate_kwargs.items()
        if k in inspect.signature(predicate).parameters
    }
    pred_result = predicate(**kwargs)
    if not pred_result:
        __handle_contract_violation(
            semantic=semantic,
            kind=kind,
            location=(
                inspect.getframeinfo(calling_frame)
                if calling_frame is not None
                else None
            ),
            kwargs=kwargs,
        )



import inspect
from types import FrameType

from pypactum._contract_violation_handler import get_contract_evaluation_semantic
from pypactum._assertion_kind import AssertionKind
from pypactum._contract_assertion_label import (
    ContractAssertionInfo,
    ContractAssertionLabel,
)
from pypactum._evaluation_semantic import EvaluationSemantic


def effective_semantic(
    parent_frame: FrameType | None,
    kind: AssertionKind,
    labels: list[ContractAssertionLabel],
) -> EvaluationSemantic:
    """Computes the effective semantic of a specific contract assertion"""

    module = inspect.getmodule(parent_frame)
    info = ContractAssertionInfo(
        kind=kind,
        module_name=module.__name__ if module is not None else "",
    )
    semantic: EvaluationSemantic = get_contract_evaluation_semantic(info)
    for label in labels:
        semantic = label(semantic, info)

    return semantic



from enum import Enum


class EvaluationSemantic(Enum):
    """The semantic with which to evaluate a contract assertion"""

    ignore = 1  # On contract violation, the contract violation handler is never called
    check = 2  # On contract violation, the contract violation handler is called, and if it returns, the program continues



import inspect
from collections.abc import Callable
from functools import wraps
from types import TracebackType
from typing import Any, Self, Literal

from pypactum._evaluation_semantic import EvaluationSemantic
from pypactum._assertion_kind import AssertionKind
from pypactum._utils._assert_contract import assert_contract
from pypactum._utils._effective_semantic import effective_semantic
from pypactum._utils._map_function_arguments import map_function_arguments
from pypactum._utils._parent_frame import get_parent_frame
from pypactum._utils._resolve_bindings import (
    resolve_bindings,
    collect_available_variables,
)
from pypactum._predicate import Predicate
from pypactum._capture_set import CaptureSet, normalize_capture_set
from pypactum._contract_assertion_label import ContractAssertionLabel


class pre:
    """Precondition assertion factory taking a predicate to evaluate on function evaluation"""

    def __init__(
        self,
        predicate: Predicate,
        /,
        *,
        capture: CaptureSet | None = None,
        clone: CaptureSet | None = None,
        labels: list[ContractAssertionLabel] | None = None,
    ):
        """Initializes the precondition assertion factory

        Keyword arguments:
            predicate: A callable evaluating the predicate to check before function evaluation.
            capture: A set of names to capture. Variables by this name can be predicate parameters.
                     Note that the wrapped function's arguments are implicitly captured.
            clone: A set of names to clone. Variables by this name can be predicate parameters.
            labels: A list of labels that determine this assertion's evaluation semantic
        """
        capture = normalize_capture_set(capture)
        clone = normalize_capture_set(clone)

        if labels is None:
            labels = []

        self.__predicate = predicate
        self.__capture = capture
        self.__clone = clone
        self.__parent_frame = get_parent_frame(inspect.currentframe())
        self.__semantic = effective_semantic(
            self.__parent_frame, AssertionKind.pre, labels
        )

    def __call__[R](
        self,
        func: Callable[..., R],
        /,
        *,
        _implicit_arg_capture: bool = True,
    ) -> Callable[..., R]:
        """Wraps the given callable in another callable that checks preconditions before executing the original callable

        Keyword arguments:
            func: Callable to wrap. Typically, a function.

        Returns:
            - `func` if the effective contract evaluation semantic is `ignore`
            - a checked wrapper compatible with `func` otherwise
        """

        if self.__semantic == EvaluationSemantic.ignore:
            return func

        @wraps(func)
        def checked_func(*args: Any, **kwargs: Any) -> R:
            sig = inspect.signature(func)
            nkwargs = map_function_arguments(sig, args, kwargs)

            # resolve bindings
            available_variables = collect_available_variables(
                self.__parent_frame,
                nkwargs,
            )
            # Implicitly capture function arguments, but explicit captures take priority
            capture = self.__capture
            if _implicit_arg_capture:
                capture = {n: n for n in sig.parameters.keys()} | capture
            resolved_kwargs = resolve_bindings(
                available_variables=available_variables,
                capture=capture,
                clone=self.__clone,
            )

            # assert precondition
            assert_contract(
                semantic=self.__semantic,
                kind=AssertionKind.pre,
                calling_frame=self.__parent_frame,
                predicate=self.__predicate,
                predicate_kwargs=resolved_kwargs,
            )

            # evaluate decorated function
            return func(*args, **kwargs)

        return checked_func

    def __enter__(self) -> Self:
        """Checks all preconditions when the scope is entered"""

        # resolve bindings
        available_variables = collect_available_variables(self.__parent_frame, {})
        resolved_kwargs = resolve_bindings(
            available_variables=available_variables,
            capture=self.__capture,
            clone=self.__clone,
        )

        # assert precondition
        assert_contract(
            semantic=self.__semantic,
            kind=AssertionKind.pre,
            calling_frame=self.__parent_frame,
            predicate=self.__predicate,
            predicate_kwargs=resolved_kwargs,
        )
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        """Doesn't do anything"""

        return False



import inspect
from types import TracebackType
from typing import Self, Literal, Any

from pypactum._pre import pre
from pypactum._post import post, PostconditionScope
from pypactum._predicate import Predicate
from pypactum._capture_set import CaptureSet, normalize_capture_set
from pypactum._contract_assertion_label import ContractAssertionLabel
from pypactum._utils._parent_frame import get_parent_frame


class invariant:
    """Invariant assertion factory taking a predicate to evaluate before and after member method evaluation"""

    __methods_to_ignore = [
        # runs before the object is initialized and therefore can't maintain any invariants
        "__new__",
        "__init_subclass__",
        "__subclasshook__",
        # required for the implementation of invariant
        "__dir__",
        "__setattr__",
        "__getattribute__",
        "__delattr__",
    ]

    def __init__(
        self,
        predicate: Predicate,
        /,
        *,
        capture: CaptureSet | None = None,
        clone: CaptureSet | None = None,
        labels: list[ContractAssertionLabel] | None = None,
    ):
        """Initializes the invariant assertion factory

        Keyword arguments:
            predicate: A callable evaluating the predicate to check before function evaluation.
            capture: A set of names to capture. Variables by this name can be predicate parameters.
                     Note that the wrapped method's arguments are implicitly captured, including `self`.
            clone: A set of names to clone. Variables by this name can be predicate parameters.
            labels: A list of labels that determine this assertion's evaluation semantic
        """
        capture = normalize_capture_set(capture)
        clone = normalize_capture_set(clone)

        if labels is None:
            labels = []

        self.__pre = pre(
            predicate,
            capture=capture,
            clone=clone,
            labels=labels,
        )
        self.__post = post(
            predicate,
            capture_after=capture,
            clone_after=clone,
            labels=labels,
            scope=PostconditionScope.Always,
        )
        # Hacky! Patch up internal parent frame to point to the correct parent, instead of this __init__ method
        parent_frame = get_parent_frame(inspect.currentframe())
        setattr(self.__pre, "_pre__parent_frame", parent_frame)
        setattr(self.__post, "_post__parent_frame", parent_frame)

    def __call__[T](self, cls: type[T], /) -> type[T]:
        """Wraps the given class that checks invariants before and after evaluation of all member methods

        Keyword arguments:
            cls: Class to wrap.

        Returns:
            - `clazz` if the current contract evaluation semantic is `ignore`
            - a checked wrapper compatible with `clazz` otherwise

        Raises:
            TypeError: if the predicate is malformed given `clazz` and the set of captured and cloned values.
        """

        def is_member_function(attribute: Any) -> bool:
            return callable(attribute) and not (
                isinstance(attribute, staticmethod)
                or isinstance(attribute, classmethod)
                or inspect.isclass(attribute)
            )

        # Copy over any inherited functions
        for name in dir(cls):
            if name not in invariant.__methods_to_ignore and name not in cls.__dict__:
                attr = getattr(cls, name)
                if is_member_function(attr):
                    setattr(cls, name, attr)

        # Decorate them
        for attr_name, attr_value in cls.__dict__.items():
            if is_member_function(attr_value):
                if attr_name in invariant.__methods_to_ignore:
                    continue
                if attr_name != "__init__":  # Can't have invariant as precondition
                    setattr(cls, attr_name, self.__pre(getattr(cls, attr_name)))
                if attr_name != "__del__":  # Can't have invariant as postcondition
                    setattr(
                        cls,
                        attr_name,
                        self.__post(
                            getattr(cls, attr_name),
                            _implicit_return_capture=False,
                            _implicit_arg_capture=True,
                        ),
                    )

        return cls

    def __enter__(self) -> Self:
        """Checks all invariants when the scope is entered

        Raises:
            TypeError: if the predicate is malformed given the set of captured and cloned values.
        """

        self.__post.__enter__()
        self.__pre.__enter__()

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        """Checks all invariants when the scope is exited"""

        self.__pre.__exit__(exc_type, exc_val, exc_tb)
        self.__post.__exit__(exc_type, exc_val, exc_tb)
        return False



import inspect

from pypactum import (
    AssertionKind,
    ContractViolation,
    EvaluationSemantic,
)


def test_contract_violation_stringification_pre_predicate_check():
    c = ContractViolation(
        comment="comment",
        kind=AssertionKind.pre,
        location=inspect.Traceback(
            filename="foo.py",
            index=0,
            function="foo",
            positions=None,
            lineno=10,
            code_context=["@pre(foobar)"],
        ),
        semantic=EvaluationSemantic.check,
        kwargs={"x": 42},
    )

    assert c.comment in str(c)
    assert f"{c.location.filename}:{c.location.lineno}" in str(c)
    assert c.location.code_context[0] in str(c)
    assert "Precondition" in str(c)
    assert "'x': 42" in str(c)


def test_contract_violation_stringification_post_predicate_check():
    c = ContractViolation(
        comment="something something",
        kind=AssertionKind.post,
        location=inspect.Traceback(
            filename="foo.py",
            index=0,
            function="foo",
            positions=None,
            lineno=42,
            code_context=["@post(foobar)"],
        ),
        semantic=EvaluationSemantic.check,
        kwargs={"foo": 42},
    )

    assert c.comment in str(c)
    assert f"{c.location.filename}:{c.location.lineno}" in str(c)
    assert c.location.code_context[0] in str(c)
    assert "Postcondition" in str(c)
    assert "'foo': 42" in str(c)
